from .cfg import *

import numpy as np
import ctypes
import sys
import os

dll_path = os.path.dirname(os.path.abspath(__file__))

rgen = ctypes.CDLL(os.path.join(dll_path, 'rgen{}'.format('.dll' if sys.platform[:3].lower() == 'win' else '.so')))
rgen.generate.argtypes = (np.ctypeslib.ndpointer(ctypes.c_longlong, flags='C_CONTIGUOUS'),
                          ctypes.POINTER(ctypes.c_int),
                          np.ctypeslib.ndpointer(ctypes.c_int, flags='C_CONTIGUOUS'),
                          ctypes.c_int, ctypes.c_int, ctypes.c_int)
rgen.generate.restype = None

class ReactionGenerator:
    def __init__(self, target, reference, vectorizer, isodesmic=False):
        self.target = target
        self.reference = reference
        self.all_species = [self.target] + [r for r in self.reference if r != self.target]

        self.vectorizer = vectorizer(self.all_species, isodesmic=isodesmic)

        self.init_energies()

    def get_basis(self):
        return self.vectorizer.basis

    def init_energies(self):
        def read_energies(entry, r):
            try:
                with open(r['energies'], 'r') as f:
                    for l in f.readlines():
                        if '=' in l:
                            l_split = [s.strip() for s in l.split('=')]
                            entry[l_split[0]] = float(l_split[-1])
            except Exception:
                print(f'Error reading file {r["energies"]}')

        def read_Hcorr(entry, r):
            try:
                with open(r['hcorr'], 'r') as f:
                    entry['Hcorr'] = [float(l.split()[-1]) for l in f
                                    if 'Thermal correction to Enthalpy' in l][0]
            except Exception:
                print(f'Error reading file {r["hcorr"]}')

        for r in self.all_species:
            entry = {}
            
            read_energies(entry, r)
            read_Hcorr(entry, r)

            if r != self.target:
                entry['dfH'] = r['dfH']
                entry['u'] = r['u']

            r['data'] = entry

    def run(self, out_dir='./out/', max_rxns=1000):
        S = np.zeros((max_rxns, len(self.get_basis())), dtype=np.int64)
        len_S = ctypes.pointer(ctypes.c_int(0))
        
        rgen.generate(S, len_S, self.get_basis(), self.vectorizer.dim, len(self.get_basis()), max_rxns)
        rxns = S[:int(len_S.contents.value)]
        if len(rxns) > 0:
            rxns = rxns[np.apply_along_axis(lambda row: abs(np.sum(row)) + np.sum(np.abs(row)), 1, rxns).argsort()]

        self.reactions = rxns
        
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)

        with open(out_dir + self.target['xyz'].split('/')[-1][:-4] + '.all', 'w') as f:
            for rxn in self.reactions:
                entry = []
                entry.append(self.vectorizer.reaction_to_str(rxn))
                entry.append(abs(sum(rxn)))
                entry.append(sum(abs(rxn)))
                entry.append(entry[1] + entry[2])

                E_sum = {}
                for r in range(len(rxn)):
                    c_frac = -rxn[r].item()/rxn[0].item()
                    E_all = self.all_species[r]['data']
                    
                    for n, e in E_all.items():
                        if n=='u':
                            E_sum[n] = E_sum.get(n, 0) + (c_frac*e)**2
                        else:
                            E_sum[n] = E_sum.get(n, 0) + c_frac*e
                
                if not 'E' in E_sum:
                    E_CV = E_sum['E(CV correction (TZ))']
                    E_IT = E_sum['E(DLPNO-CCSD(T)/TZ-IT)']-E_sum['E(DLPNO-CCSD(T)/CC-PVTZ)']
                    E_SO = E_sum['E(DKH correction)']
                    E_r = E_sum['E(DLPNO-CCSD(T)/CBS)'] + E_CV + E_IT + E_SO
                    E_r_HF = E_sum['E(HF/CBS)'] + E_SO
                    E_corr_r = E_r - E_r_HF
                else:
                    E_CV = 0
                    E_IT = 0
                    E_SO = 0
                    E_r = E_sum['E']
                    E_r_HF = 0
                    E_corr_r = E_r
                    
                
                H_r = E_r + E_sum['Hcorr']
                H_f = E_sum['dfH'] - H_r*eh_to_cal

                u = E_sum['u']**0.5

                entry.append(round(H_f, 3))
                entry.append(round(E_corr_r*eh_to_cal, 3))
                entry.append(round(E_sum['Hcorr']*eh_to_cal, 3))
                entry.append(round(entry[-1]+entry[-2], 3))
                entry.append(round(u, 3))
                entry.append(round(2*u, 3))
                entry.append(round(H_r*eh_to_cal, 3))
                entry.append(0)
                entry.append(round(E_CV*eh_to_cal, 3))
                entry.append(round(E_IT*eh_to_cal, 3))
                entry.append(0)
                entry.append(round(E_SO*eh_to_cal, 3))

                entry.append(self.vectorizer.reaction_to_names(rxn))

                f.write(';'.join(map(str, entry)) + '\n')



from .cfg import *

import numpy as np
import scipy as sp

class EnthalpyEstimator:
    def __init__(self, target, reference, vectorizer, isodesmic=False):
        self.target = target
        self.reference = reference
        self.all_species = [self.target] + [r for r in self.reference if r != self.target]
        
        self.excluded = []
        self.reaction = None
        self.directions = None
        self.dfH_grad = None
        self.u_grad = None

        self.vectorizer = vectorizer(self.all_species, isodesmic=isodesmic)

        self.init_dfH()

    def get_basis(self):
        return self.vectorizer.basis

    def init_dfH(self):
        def read_energies(entry, r):
            try:
                with open(r['energies'], 'r') as f:
                    for l in f.readlines()[2:]:
                        if '=' in l:
                            l_split = [s.strip() for s in l.split('=')]
                            entry[l_split[0]] = float(l_split[-1])
            except Exception:
                print(f'Error reading file {r["energies"]}')

        def read_Hcorr(entry, r):
            try:
                with open(r['hcorr'], 'r') as f:
                    entry['Hcorr'] = [float(l.split()[-1]) for l in f
                                    if 'Thermal correction to Enthalpy' in l][0]
            except Exception:
                print(f'Error reading file {r["hcorr"]}')

        def calc_drH(E_all):
            if not 'E' in E_all:
                E_CV = E_all['E(CV correction (TZ))']
                E_IT = E_all['E(DLPNO-CCSD(T)/TZ-IT)']-E_all['E(DLPNO-CCSD(T)/CC-PVTZ)']
                E_SO = E_all['E(DKH correction)']
                E_r = E_all['E(DLPNO-CCSD(T)/CBS)'] + E_CV + E_IT + E_SO
            else:
                E_r = E_all['E']
            H_r = (E_r + E_all['Hcorr'])*eh_to_cal
            
            return H_r
        
        self.drH = []
        self.dfH = []
        self.u = []
        for r in self.all_species:
            E_all = {}
            read_energies(E_all, r)
            read_Hcorr(E_all, r)

            drH = calc_drH(E_all)
            self.drH.append(drH)
            
            if r != self.target:
                self.dfH.append(r['dfH'] - drH)
                self.u.append(r['u']**2)
            else:
                self.dfH.append(-drH)
                self.u.append(0)

        self.drH = np.array(self.drH)
        self.dfH = np.array(self.dfH)
        self.u = np.array(self.u)
    
    def run(self, optimize=False, minimize=True, dfH_thres=0.5, min_ref=8):
        dfH, u = self.calc(minimize=minimize)

        if optimize:
            to_exclude = []
            while (np.linalg.norm(self.dfH_grad) > dfH_thres 
                   and len(self.get_basis()) - len(to_exclude) > min_ref):
                to_exclude.append(np.argmax(np.abs(self.dfH_grad[1:])))
                to_exclude.sort()
                
                dfH, u = self.calc(to_exclude, minimize=minimize)
            
            self.excluded = to_exclude
            
        return dfH, u
    
    def calc(self, to_exclude=[], minimize=False):
        A = np.delete(self.get_basis()[1:].T, to_exclude, axis=1)
        b = self.get_basis()[0]
        NS = sp.linalg.null_space(A)
        
        if minimize:
            U = np.diag(np.delete(self.u[1:], to_exclude))
            M = np.block([[U, A.T], [A, np.zeros((A.shape[0], A.shape[0]))]])
            Mb = np.r_[np.zeros(A.shape[1]), b]
            x0 = np.linalg.lstsq(M, Mb, rcond=-1)[0][:A.shape[1]]
        else:
            x0 = np.linalg.lstsq(A, b, rcond=-1)[0]
        
        rxn = np.r_[-1, self._fill_zeros(x0, to_exclude)]
        dirs = np.r_[np.zeros(NS.shape[1])[None], self._fill_zeros(NS, to_exclude)]
        dfH, self.dfH_grad = self.evaluate(self.dfH, rxn, dirs)
        u = rxn.T@np.diag(self.u)@rxn
        
        self.reaction = rxn
        self.directions = dirs
        
        return dfH, u
    
    def evaluate(self, func, x0, dirs):
        y = func.dot(x0)
        grad = dirs.dot(np.linalg.inv(dirs.T.dot(dirs))).dot(dirs.T).dot(func) # projection onto subspace
        return y, grad
    
    @staticmethod
    def _fill_zeros(array, at, axis=0):
        if len(at) > 0:
            return np.insert(array, np.array(at) - np.arange(len(at)), 0, axis)
        return array



# read version from installed package
from importlib.metadata import version
__version__ = version("enthalpy_estimator")

def mol_template(xyz, energies, hcorr='', name='', smiles='', dfH=0, u=0):
    return {
        'name': name,
        'smiles': smiles,
        'dfH': dfH,
        'u': u,
        'xyz': xyz,
        'energies': energies,
        'hcorr': hcorr,
    }


from .simple_vectorizer import MolVectorizer

from rdkit import Chem
from rdkit.Chem import AllChem
from rdkit.Chem import rdqueries
import numpy as np

class SMILESVectorizer(MolVectorizer):
    def __init__(self, all_species, isodesmic=False):
        self.all_species = all_species
        self.isodesmic = isodesmic

        self.init_basis()

    @staticmethod
    def cleanup_xyz(filename):
        ind = open(filename).readlines()
        if ' ' in ind[0]:
            n_atoms = str(len(ind))
            ind = [n_atoms + '\n\n'] + ind

        return ''.join(ind)

    @staticmethod
    def smiles_to_mol(smi):
        return AllChem.AddHs(Chem.MolFromSmiles(smi))

    def mol_to_vec(self, mol):
        r_vec = [0]*self.dim

        for i, el in enumerate(self.elements):
            q = rdqueries.AtomNumEqualsQueryAtom(int(el))
            r_vec[i] = len(mol.GetAtomsMatchingQuery(q))

        if self.isodesmic:
            for i, b in enumerate(self.bonds):
                r_vec[len(self.elements)+i] = len(mol.GetSubstructMatches(b))

        return r_vec

    def init_basis(self):
        self.mols = [self.smiles_to_mol(r['smiles']) for r in self.all_species]
        self.elements = np.array(sorted(
            {a.GetAtomicNum() for mol in self.mols for a in mol.GetAtoms()}, reverse=True))
        self.bonds = [Chem.MolFromSmarts(s) for s in (
            '[#6]-[#1]', '[#6]-[#6]', '[#6]=[#6]', '[#6]#[#6]', '[#6]:[#6]')]
        self.symbols = np.array(
            [Chem.GetPeriodicTable().GetElementSymbol(int(e)) for e in self.elements])

        self.dim = len(self.symbols) + \
            (len(self.bonds) if self.isodesmic else 0)

        self.basis = np.array([self.mol_to_vec(mol) for mol in self.mols], dtype=np.int32)




import numpy as np

class MolVectorizer:
    def __init__(self, all_species, isodesmic=False):
        self.all_species = all_species

        if isodesmic:
            print('Warning: simple vectorizer does not support isodesmic reactions')
        
        self.init_basis()
    
    def read_atoms(self, r):
        try:
            with open(r['xyz'], 'r') as f:
                atoms = {}
                for l in f:
                    if ' ' not in l: continue
                    a = l.split()[0].strip()
                    atoms[a] = atoms.get(a, 0) + 1

                return atoms
        except OSError:
            print(f'Error reading file {r["xyz"]}')
            return {}

    def init_basis(self):
        self.mols = [self.read_atoms(r) for r in self.all_species]
        self.symbols = np.array(sorted({el for atoms in self.mols for el in atoms.keys()}))
        self.dim = len(self.symbols)
        self.basis = np.array([[atoms.get(e, 0) for e in self.symbols] for atoms in self.mols], dtype=np.int32)
    
    @staticmethod
    def coef_to_str(c):
        return str(c) if c!=1 else ''

    def vector_to_str(self, v):
        els = np.where(v[:len(self.symbols)] != 0)
        el_n = np.column_stack([self.symbols[els], v[els]])
        el_n[el_n[:, 1] == '1', 1] = ''
        s = ''.join(''.join(el_n.flatten()))
                    
        return s
    
    def reaction_to_str(self, rxn):
        r = rxn * np.sign(rxn[0])
        reagents = np.where(r > 0)[0]
        products = np.where(r < 0)[0]

        s = ''
        s += '+'.join(
                ''.join([self.coef_to_str(r[i].item()), self.vector_to_str(self.basis[i])]) 
                for i in reagents)
        s += '->'
        s += '+'.join(
                ''.join([self.coef_to_str(-r[i].item()), self.vector_to_str(self.basis[i])]) 
                for i in products)

        return s
    
    def reaction_to_names(self, rxn):
        r = rxn * np.sign(rxn[0])
        reagents = np.where(r > 0)[0]
        products = np.where(r < 0)[0]

        s = ''
        s += '+'.join(
                ''.join([str(r[i].item()) + '*'
                         if abs(r[i].item()) != 1
                         else '', 
                         self.all_species[i]['name']
                         if 'name' in self.all_species[i] and self.all_species[i]['name'] is not ''
                         else self.vector_to_str(self.basis[i])]) 
                for i in reagents)
        s += '->'
        s += '+'.join(
                ''.join([str(-r[i].item()) + '*'
                         if abs(r[i].item()) != 1
                         else '', 
                         self.all_species[i]['name']
                         if 'name' in self.all_species[i] and self.all_species[i]['name'] is not ''
                         else self.vector_to_str(self.basis[i])]) 
                for i in products)

        return s